@@ -32,6 +32,7 @@
 #include <glib.h>
 #include <gdk/gdkkeysyms.h>
 #include <glib/gslist.h>
+#include <gtk-layer-shell/gtk-layer-shell.h>
 
 #include <lightdm.h>
 
@@ -69,12 +70,13 @@ static gchar *current_language;
 /* Screensaver values */
 int timeout, interval, prefer_blanking, allow_exposures;
 
-static GdkColor *default_background_color = NULL;
+static GdkRGBA *default_background_color = NULL;
 static gboolean cancelling = FALSE, prompted = FALSE;
 static gboolean prompt_active = FALSE, password_prompted = FALSE;
-static GdkRegion *window_region = NULL;
 static gchar *wp_mode = NULL;
 
+static gboolean wayland = FALSE;
+
 typedef struct
 {
   gboolean is_prompt;
@@ -244,7 +246,7 @@ is_valid_session (GList* items, const gchar* session)
 static gchar *
 get_session (void)
 {
-    return g_strdup (current_session);
+    return g_strdup (lightdm_greeter_get_default_session_hint (greeter));
 }
 
 static void
@@ -486,92 +488,34 @@ get_absolute_position (const DimensionPosition *p, gint screen, gint window)
 static void
 center_window (GtkWindow *window, GtkAllocation *unused, const WindowPosition *pos)
 {   
-    GdkScreen *screen = gtk_window_get_screen (window);
     GtkAllocation allocation;
     GdkRectangle monitor_geometry;
+    if (wayland) return;
 
-    gdk_screen_get_monitor_geometry (screen, gdk_screen_get_primary_monitor (screen), &monitor_geometry);
+    gdk_monitor_get_geometry (gdk_display_get_primary_monitor (gdk_display_get_default ()), &monitor_geometry);
     gtk_widget_get_allocation (GTK_WIDGET (window), &allocation);
     gtk_window_move (window,
                      monitor_geometry.x + get_absolute_position (&pos->x, monitor_geometry.width, allocation.width),
                      monitor_geometry.y + get_absolute_position (&pos->y, monitor_geometry.height, allocation.height));
 }
 
-static GdkRegion *
-cairo_region_from_rectangle (gint width, gint height, gint radius)
-{
-    GdkRegion *region;
-
-    gint x = radius, y = 0;
-    gint xChange = 1 - (radius << 1);
-    gint yChange = 0;
-    gint radiusError = 0;
-
-    GdkRectangle rect;
-
-    rect.x = radius;
-    rect.y = radius;
-    rect.width = width - radius * 2;
-    rect.height = height - radius * 2;
-
-    region = gdk_region_rectangle (&rect);
-
-    while(x >= y)
-    {
-
-        rect.x = -x + radius;
-        rect.y = -y + radius;
-        rect.width = x - radius + width - rect.x;
-        rect.height =  y - radius + height - rect.y;
-
-        gdk_region_union_with_rect(region, &rect);
-
-        rect.x = -y + radius;
-        rect.y = -x + radius;
-        rect.width = y - radius + width - rect.x;
-        rect.height =  x - radius + height - rect.y;
-
-        gdk_region_union_with_rect(region, &rect);
-
-        y++;
-        radiusError += yChange;
-        yChange += 2;
-        if(((radiusError << 1) + xChange) > 0)
-        {
-            x--;
-            radiusError += xChange;
-            xChange += 2;
-        }
-   }
-
-   return region;
-}
-
+/* Use the much simpler fake transparency by drawing the window background with Cairo for Gtk3 */
 static gboolean
-login_window_size_allocate (GtkWidget *widget, GdkRectangle *allocation, gpointer user_data)
+login_window_draw (GtkWidget *widget, cairo_t *cr, gpointer user_data)
 {
-    gint    radius = 2;
+    gdk_cairo_set_source_rgba (cr, default_background_color);
+    cairo_paint (cr);
 
-    GdkWindow *window = gtk_widget_get_window (widget);
-    if (window_region)
-        gdk_region_destroy(window_region);
-    window_region = cairo_region_from_rectangle (allocation->width, allocation->height, radius);
-    if (window) {
-        gdk_window_shape_combine_region(window, window_region, 0, 0);
-        gdk_window_input_shape_combine_region(window, window_region, 0, 0);
-    }
-
-    return TRUE;
+    return FALSE;
 }
 
-
 static void draw_background (cairo_t *c, GdkPixbuf *bg, gint m_width, gint m_height)
 {
     GdkPixbuf *p = NULL;
     gint p_height, p_width, offset_x = 0, offset_y = 0;
     gdouble scale_x, scale_y;
 
-    gdk_cairo_set_source_color (c, default_background_color);
+    gdk_cairo_set_source_rgba (c, default_background_color);
     cairo_rectangle (c, 0, 0, m_width, m_height);
     cairo_fill (c);
 
@@ -613,12 +557,10 @@ static void draw_background (cairo_t *c, GdkPixbuf *bg, gint m_width, gint m_hei
 }
 
 static gboolean
-background_window_expose (GtkWidget    *widget,
-                                       GdkEventExpose *event,
-                                       gpointer user_data)
+background_window_draw (GtkWidget *widget, cairo_t *cr, gpointer user_data)
 {
     GdkWindow *wd = gtk_widget_get_window (widget);
-    cairo_t *cr = gdk_cairo_create (wd);
+
     draw_background (cr, default_background_pixbuf, gdk_window_get_width (wd), gdk_window_get_height (wd));
     return FALSE;
 }
@@ -942,7 +884,7 @@ static void set_displayed_user (LightDMGreeter *greeter, gchar *username)
     }
 
     set_login_button_label (greeter, username);
-    set_user_background (username);
+    //set_user_background (username);
     set_user_image (username);
     user = lightdm_user_list_get_user_by_name (lightdm_user_list_get_instance (), username);
     if (user)
@@ -1060,7 +1002,7 @@ void
 login_cb (GtkWidget *widget)
 {
     /* Reset to default screensaver values */
-    if (lightdm_greeter_get_lock_hint (greeter))
+    if (!wayland && lightdm_greeter_get_lock_hint (greeter))
         XSetScreenSaver(gdk_x11_display_get_xdisplay(gdk_display_get_default ()), timeout, interval, prefer_blanking, allow_exposures);        
 
     gtk_widget_set_sensitive (GTK_WIDGET (username_entry), FALSE);
@@ -1380,6 +1322,42 @@ load_user_list (void)
     g_free (last_user);
 }
 
+static int screen_width (GdkScreen *scr)
+{
+    GdkDisplay *disp = gdk_screen_get_display (scr == NULL ? gdk_screen_get_default () : scr);
+    GdkMonitor *mon;
+    GdkRectangle rect;
+    int min = -1, max = 0, i;
+
+    for (i = 0; i < gdk_display_get_n_monitors (disp); i++)
+    {
+        mon = gdk_display_get_monitor (disp, i);
+        gdk_monitor_get_geometry (mon, &rect);
+        if (min == -1 || rect.x < min) min = rect.x;
+        if (rect.x + rect.width > max) max = rect.x + rect.width;
+    }
+
+    return max - min;
+}
+
+static int screen_height (GdkScreen *scr)
+{
+    GdkDisplay *disp = gdk_screen_get_display (scr == NULL ? gdk_screen_get_default () : scr);
+    GdkMonitor *mon;
+    GdkRectangle rect;
+    int min = -1, max = 0, i;
+
+    for (i = 0; i < gdk_display_get_n_monitors (disp); i++)
+    {
+        mon = gdk_display_get_monitor (disp, i);
+        gdk_monitor_get_geometry (mon, &rect);
+        if (min == -1 || rect.y < min) min = rect.y;
+        if (rect.y + rect.height > max) max = rect.y + rect.height;
+    }
+
+    return max - min;
+}
+
 /* The following code for setting a RetainPermanent background pixmap was taken
    originally from Gnome, with some fixes from MATE. see:
    https://github.com/mate-desktop/mate-desktop/blob/master/libmate-desktop/mate-bg.c */
@@ -1391,12 +1369,12 @@ create_root_surface (GdkScreen *screen)
     Pixmap pixmap;
     cairo_surface_t *surface;
 
-    number = gdk_screen_get_number (screen);
-    width = gdk_screen_get_width (screen);
-    height = gdk_screen_get_height (screen);
+    number = 0;
+    width = screen_width (screen);
+    height = screen_height (screen);
 
     /* Open a new connection so with Retain Permanent so the pixmap remains when the greeter quits */
-    gdk_flush ();
+    gdk_display_flush (gdk_display_get_default ());
     display = XOpenDisplay (gdk_display_get_name (gdk_screen_get_display (screen)));
     if (!display)
     {
@@ -1424,7 +1402,7 @@ set_root_pixmap_id (GdkScreen *screen,
                          Display *display,
                          Pixmap xpixmap)
 {
-    Window xroot = RootWindow (display, gdk_screen_get_number (screen));
+    Window xroot = RootWindow (display, 0);
     char *atom_names[] = {"_XROOTPMAP_ID", "ESETROOT_PMAP_ID"};
     Atom atoms[G_N_ELEMENTS(atom_names)] = {0};
 
@@ -1453,7 +1431,7 @@ set_root_pixmap_id (GdkScreen *screen,
                 XFree (data_root);
                 XFree (data_esetroot);
 
-                gdk_error_trap_push ();
+                gdk_x11_display_error_trap_push (gdk_x11_lookup_xdisplay (display));
                 if (xrootpmap && xrootpmap == esetrootpmap) {
                     XKillClient (display, xrootpmap);
                 }
@@ -1462,7 +1440,7 @@ set_root_pixmap_id (GdkScreen *screen,
                 }
 
                 XSync (display, False);
-                gdk_error_trap_pop ();
+                gdk_x11_display_error_trap_pop (gdk_x11_lookup_xdisplay (display));
 
             }
         }
@@ -1508,7 +1486,7 @@ set_surface_as_root (GdkScreen *screen, cairo_surface_t *surface)
      */
     Display *display = GDK_DISPLAY_XDISPLAY (gdk_screen_get_display (screen));
     Pixmap pixmap_id = cairo_xlib_surface_get_drawable (surface);
-    Window xroot = RootWindow (display, gdk_screen_get_number (screen));
+    Window xroot = RootWindow (display, 0);
 
     XGrabServer (display);
 
@@ -1526,44 +1504,37 @@ set_background (GdkPixbuf *new_bg)
     GdkRectangle monitor_geometry;
     GdkPixbuf *bg = NULL;
     GSList *iter;
-    gint i, num_screens = 1;
 
     if (new_bg)
         bg = new_bg;
     else
         bg = default_background_pixbuf;
 
-    #if GDK_VERSION_CUR_STABLE < G_ENCODE_VERSION(3, 10)
-        num_screens = gdk_display_get_n_screens (gdk_display_get_default ());
-    #endif
-
     /* Set the background */
-    for (i = 0; i < num_screens; i++)
+    GdkScreen *screen;
+    cairo_surface_t *surface;
+    cairo_t *c;
+    gint monitor;
+
+    screen = gdk_display_get_default_screen (gdk_display_get_default ());
+    surface = create_root_surface (screen);
+    c = cairo_create (surface);
+
+    for (monitor = 0; monitor < gdk_display_get_n_monitors (gdk_display_get_default ()); monitor++)
     {
-        GdkScreen *screen;
-        cairo_surface_t *surface;
-        cairo_t *c;
-        gint monitor;
-
-        screen = gdk_display_get_screen (gdk_display_get_default (), i);
-        surface = create_root_surface (screen);
-        c = cairo_create (surface);
-
-        for (monitor = 0; monitor < gdk_screen_get_n_monitors (screen); monitor++)
-        {
-            gdk_screen_get_monitor_geometry (screen, monitor, &monitor_geometry);
-            draw_background (c, bg, monitor_geometry.width, monitor_geometry.height);
-            iter = g_slist_nth (backgrounds, monitor);
-            gtk_widget_queue_draw (GTK_WIDGET (iter->data));
-        }
-
-        cairo_destroy (c);
-
-        /* Refresh background */
-        gdk_flush ();
-        set_surface_as_root (screen, surface);
-        cairo_surface_destroy (surface);
+        gdk_monitor_get_geometry (gdk_display_get_monitor (gdk_display_get_default (), monitor), &monitor_geometry);
+        draw_background (c, bg, monitor_geometry.width, monitor_geometry.height);
+        iter = g_slist_nth (backgrounds, monitor);
+        gtk_widget_queue_draw (GTK_WIDGET (iter->data));
     }
+
+    cairo_destroy (c);
+
+    /* Refresh background */
+    gdk_display_flush (gdk_display_get_default ());
+    set_surface_as_root (screen, surface);
+    cairo_surface_destroy (surface);
+
     gtk_widget_queue_draw (GTK_WIDGET (login_window));
 }
 
@@ -1611,7 +1582,7 @@ focus_upon_map (GdkXEvent *gxevent, GdkEvent *event, gpointer  data)
 
         /* Check to see if this window is our onboard window, since we don't want to focus it. */
         if (keyboard_win)
-                keyboard_xid = gdk_x11_drawable_get_xid (keyboard_win);
+                keyboard_xid = gdk_x11_window_get_xid (keyboard_win);
 
         if (xwin != keyboard_xid
             && win_type != GDK_WINDOW_TYPE_HINT_TOOLTIP
@@ -1653,15 +1624,13 @@ main (int argc, char **argv)
     GdkRectangle monitor_geometry;
     GtkBuilder *builder;
     GtkCellRenderer *renderer;
-    GtkWidget *infobar_compat, *content_area;
     gchar *value, *state_dir;
-    GdkColor background_color;
+    GdkRGBA background_color;
     GError *error = NULL;
 
     /* Background windows */
-    gint monitor, scr;
-    gint numScreens = 1;
-    GdkScreen *screen;
+    gint monitor;
+    GSList *iter;
     GtkWidget *window;
 
     Display* display;
@@ -1670,6 +1639,8 @@ main (int argc, char **argv)
     GPid indicator_pid = 0, spi_pid = 0;
     #endif
 
+    if (getenv ("WAYLAND_DISPLAY")) wayland = TRUE;
+
     /* Prevent memory from being swapped out, as we are dealing with passwords */
     mlockall (MCL_CURRENT | MCL_FUTURE);
 
@@ -1684,9 +1655,6 @@ main (int argc, char **argv)
 
     g_unix_signal_add(SIGTERM, (GSourceFunc)gtk_main_quit, NULL);
 
-    /* init threads */
-    gdk_threads_init();
-
     /* init gtk */
     gtk_init (&argc, &argv);
     
@@ -1716,14 +1684,14 @@ main (int argc, char **argv)
         return EXIT_FAILURE;
 
     /* Set default cursor */
-    gdk_window_set_cursor (gdk_get_default_root_window (), gdk_cursor_new (GDK_LEFT_PTR));
+    gdk_window_set_cursor (gdk_get_default_root_window (), gdk_cursor_new_for_display (gdk_display_get_default (), GDK_LEFT_PTR));
 
     /* Get background colour */
     value = g_key_file_get_value (config, "greeter", "desktop_bg", NULL);
-    if (!value || !gdk_color_parse (value, &background_color))
-            gdk_color_parse ("#C0C0C0", &background_color);
+    if (!value || !gdk_rgba_parse (&background_color, value))
+            gdk_rgba_parse (&background_color, "#C0C0C0");
     if (value) g_free (value);
-    default_background_color = gdk_color_copy (&background_color);
+    default_background_color = gdk_rgba_copy (&background_color);
 
     /* Get background image */
     value = g_key_file_get_value (config, "greeter", "wallpaper", NULL);
@@ -1780,7 +1748,7 @@ main (int argc, char **argv)
     g_free (value);
 
     display = gdk_x11_display_get_xdisplay(gdk_display_get_default ());
-    if (lightdm_greeter_get_lock_hint (greeter)) {
+    if (!wayland && lightdm_greeter_get_lock_hint (greeter)) {
         XGetScreenSaver(display, &timeout, &interval, &prefer_blanking, &allow_exposures);
         XForceScreenSaver(display, ScreenSaverActive);
         XSetScreenSaver(display, screensaver_timeout, 0, ScreenSaverActive, DefaultExposures);
@@ -1795,33 +1763,13 @@ main (int argc, char **argv)
     user_combo = GTK_COMBO_BOX (gtk_builder_get_object (builder, "user_combobox"));
     username_entry = GTK_ENTRY (gtk_builder_get_object (builder, "username_entry"));
     password_entry = GTK_ENTRY (gtk_builder_get_object (builder, "password_entry"));
-
-    /* Add InfoBar via code for GTK+2 compatability */
-    infobar_compat = GTK_WIDGET(gtk_builder_get_object(builder, "infobar_compat"));
-    info_bar = GTK_INFO_BAR (gtk_info_bar_new());
-    gtk_info_bar_set_message_type(info_bar, GTK_MESSAGE_ERROR);
-    gtk_widget_set_name(GTK_WIDGET(info_bar), "greeter_infobar");
-    content_area = gtk_info_bar_get_content_area(info_bar);
-
+    info_bar = GTK_INFO_BAR (GTK_WIDGET(gtk_builder_get_object(builder, "greeter_infobar")));
     message_label = GTK_LABEL (gtk_builder_get_object (builder, "message_label"));
-    g_object_ref(message_label);
-    gtk_container_remove(GTK_CONTAINER(infobar_compat), GTK_WIDGET(message_label));
-    gtk_container_add(GTK_CONTAINER(content_area), GTK_WIDGET(message_label));
-    g_object_unref(message_label);
-
-    gtk_container_add(GTK_CONTAINER(infobar_compat), GTK_WIDGET(info_bar));
-
     cancel_button = GTK_BUTTON (gtk_builder_get_object (builder, "cancel_button"));
     login_button = GTK_BUTTON (gtk_builder_get_object (builder, "login_button"));
 
-    g_signal_connect (G_OBJECT (login_window), "size-allocate", G_CALLBACK (login_window_size_allocate), NULL);
-
-    /* To maintain compatability with GTK+2, set special properties here */
-    gtk_container_set_border_width (GTK_CONTAINER(gtk_builder_get_object (builder, "vbox2")), 18);
-    gtk_container_set_border_width (GTK_CONTAINER(gtk_builder_get_object (builder, "content_frame")), 14);
-    gtk_container_set_border_width (GTK_CONTAINER(gtk_builder_get_object (builder, "buttonbox_frame")), 8);
-    gtk_widget_set_tooltip_text(GTK_WIDGET(password_entry), _("Enter your password"));
-    gtk_widget_set_tooltip_text(GTK_WIDGET(username_entry), _("Enter your username"));
+    //g_signal_connect (G_OBJECT (login_window), "draw", G_CALLBACK (login_window_draw), NULL);
+    //gtk_widget_queue_draw (GTK_WIDGET (login_window));
 
 #ifdef START_INDICATOR_SERVICES
     init_indicators (config, &indicator_pid, &spi_pid);
@@ -1852,46 +1800,41 @@ main (int argc, char **argv)
     gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (user_combo), renderer, "text", 1);
     //gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (user_combo), renderer, "weight", 2);
 
-    #if GDK_VERSION_CUR_STABLE < G_ENCODE_VERSION(3, 10)
-        numScreens = gdk_display_get_n_screens (gdk_display_get_default());
-    #endif
-
     /* Set up the background images */	
-    gdk_color_parse ("#000000", &background_color);
-    for (scr = 0; scr < numScreens; scr++)
+    for (monitor = 0; monitor < gdk_display_get_n_monitors (gdk_display_get_default ()); monitor++)
     {
-        screen = gdk_display_get_screen (gdk_display_get_default (), scr);
-        for (monitor = 0; monitor < gdk_screen_get_n_monitors (screen); monitor++)
-        {
-            gdk_screen_get_monitor_geometry (screen, monitor, &monitor_geometry);
-        
-            window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
-            gtk_window_set_type_hint(GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_DESKTOP);
-            gtk_widget_modify_bg(GTK_WIDGET(window), GTK_STATE_NORMAL, &background_color);
-            gtk_window_set_screen(GTK_WINDOW(window), screen);
-            gtk_window_set_keep_below(GTK_WINDOW(window), TRUE);
-            gtk_widget_set_size_request(window, monitor_geometry.width, monitor_geometry.height);
-            gtk_window_set_resizable (GTK_WINDOW(window), FALSE);
-            gtk_widget_set_app_paintable (GTK_WIDGET(window), TRUE);
-            gtk_window_move (GTK_WINDOW(window), monitor_geometry.x, monitor_geometry.y);
+        gdk_monitor_get_geometry (gdk_display_get_monitor (gdk_display_get_default (), monitor), &monitor_geometry);
 
-            backgrounds = g_slist_prepend(backgrounds, window);
-            gtk_widget_show (window);
-            g_signal_connect (G_OBJECT (window), "expose-event", G_CALLBACK (background_window_expose), NULL);
-            gtk_widget_queue_draw (GTK_WIDGET(window));
+        window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
+        gtk_window_set_type_hint(GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_DESKTOP);
+        gtk_window_set_keep_below(GTK_WINDOW(window), TRUE);
+        if (wayland) gtk_window_set_default_size (GTK_WINDOW (window), monitor_geometry.width, monitor_geometry.height);
+        else gtk_widget_set_size_request(window, monitor_geometry.width, monitor_geometry.height);
+        gtk_window_set_resizable (GTK_WINDOW(window), FALSE);
+        gtk_widget_set_app_paintable (GTK_WIDGET(window), TRUE);
+        if (!wayland) gtk_window_move (GTK_WINDOW(window), monitor_geometry.x, monitor_geometry.y);
+        else
+        {
+            gtk_layer_init_for_window (GTK_WINDOW (window));
+            gtk_layer_set_layer (GTK_WINDOW (window), GTK_LAYER_SHELL_LAYER_BACKGROUND);
+            gtk_layer_set_monitor (GTK_WINDOW (window), gdk_display_get_monitor (gdk_display_get_default (), monitor));
+            gtk_layer_set_anchor (GTK_WINDOW (window), GTK_LAYER_SHELL_EDGE_TOP, TRUE);
+            gtk_layer_set_anchor (GTK_WINDOW (window), GTK_LAYER_SHELL_EDGE_BOTTOM, TRUE);
+            gtk_layer_set_anchor (GTK_WINDOW (window), GTK_LAYER_SHELL_EDGE_LEFT, TRUE);
+            gtk_layer_set_anchor (GTK_WINDOW (window), GTK_LAYER_SHELL_EDGE_RIGHT, TRUE);
         }
+
+        backgrounds = g_slist_prepend(backgrounds, window);
     }
     backgrounds = g_slist_reverse(backgrounds);
 
     if (lightdm_greeter_get_hide_users_hint (greeter))
     {
-        /* Set the background to default */
-        set_background (NULL);
+        //set_background (NULL);
         start_authentication ("*other");
     }
     else
     {
-        /* This also sets the background to user's */
         load_user_list ();
         gtk_widget_show (GTK_WIDGET (cancel_button));
         gtk_widget_show (GTK_WIDGET (user_combo));
@@ -1917,22 +1860,43 @@ main (int argc, char **argv)
     }
 
     gtk_builder_connect_signals(builder, greeter);
+    if (wayland)
+    {
+        gtk_layer_init_for_window (GTK_WINDOW (login_window));
+        gtk_layer_set_layer (GTK_WINDOW (login_window), GTK_LAYER_SHELL_LAYER_TOP);
+        gtk_layer_set_monitor (GTK_WINDOW (login_window), gdk_display_get_monitor (gdk_display_get_default (), 0));
+        gtk_layer_set_keyboard_interactivity (GTK_WINDOW (login_window), TRUE);
+        gtk_window_set_decorated (login_window, TRUE);
+    }
 
     gtk_widget_show (GTK_WIDGET (login_window));
     center_window (login_window,  NULL, &main_window_pos);
     g_signal_connect (GTK_WIDGET (login_window), "size-allocate", G_CALLBACK (center_window), &main_window_pos);
 
     gtk_widget_show (GTK_WIDGET (login_window));
-    gdk_window_focus (gtk_widget_get_window (GTK_WIDGET (login_window)), GDK_CURRENT_TIME);
+    gtk_widget_hide (GTK_WIDGET (info_bar));
 
-    /* focus fix (source: unity-greeter) */
-    GdkWindow* root_window = gdk_get_default_root_window ();
-    gdk_window_set_events (root_window, gdk_window_get_events (root_window) | GDK_SUBSTRUCTURE_MASK);
-    gdk_window_add_filter (root_window, focus_upon_map, NULL);
+    for (monitor = 0; monitor < gdk_display_get_n_monitors (gdk_display_get_default ()); monitor++)
+    {
+        iter = g_slist_nth (backgrounds, monitor);
+        window = GTK_WIDGET (iter->data);
+        gtk_widget_show (window);
+        gtk_window_set_decorated (GTK_WINDOW (window), FALSE);
+        g_signal_connect (G_OBJECT (window), "draw", G_CALLBACK (background_window_draw), NULL);
+        gtk_widget_queue_draw (GTK_WIDGET(window));
+    }
+
+    if (!wayland)
+	{
+		gdk_window_focus (gtk_widget_get_window (GTK_WIDGET (login_window)), GDK_CURRENT_TIME);
+
+		/* focus fix (source: unity-greeter) */
+		GdkWindow* root_window = gdk_get_default_root_window ();
+		gdk_window_set_events (root_window, gdk_window_get_events (root_window) | GDK_SUBSTRUCTURE_MASK);
+		gdk_window_add_filter (root_window, focus_upon_map, NULL);
+	}
 
-    gdk_threads_enter();
     gtk_main ();
-    gdk_threads_leave();
 
 #ifdef START_INDICATOR_SERVICES
     if (indicator_pid)
@@ -1951,8 +1915,9 @@ main (int argc, char **argv)
     if (default_background_pixbuf)
         g_object_unref (default_background_pixbuf);
     if (default_background_color)
-        gdk_color_free (default_background_color);
+        gdk_rgba_free (default_background_color);
 
+    if (!wayland)
     {
 	int screen = XDefaultScreen (display);
 	Window w = RootWindow (display, screen);
